/********************************************************************
 * This file Copyright, 2005 Michael D'Angelo, PhoeniX WorX Online Gaming
 * It is licensed under the modified BSD License, the rest of the program is GPL
 * ********************************************************************/
using System;
using System.Net;

namespace Phoenix
{
	public sealed class ByteConverter
	{
		// instances not creatable
		ByteConverter()
		{
		}

		/// <summary>
		/// Holds value of whether environment is Big or Little endian
		/// </summary>
		/// <value>!ByteConverter.IsLittleEndian</value>
		public static readonly bool IsBigEndian = !BitConverter.IsLittleEndian; // is there another way to do this?

		public static unsafe int IndexOf(byte b, byte[] buffer, int startIndex)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (startIndex < 0 || startIndex > buffer.Length)
				throw new IndexOutOfRangeException();

			int length = buffer.Length;
			fixed(byte* arr = buffer)
			{
				byte* aptr = arr + startIndex;
				for(int i=startIndex;i<length;i++)
				{
					//if(aptr->Equals(b))
					if((*aptr) == b)
						return i;
					aptr++;
				}
			}
			return -1;
		}

		/// <summary>
		/// Find the index of the specified value's instanceNum instance.  
		/// </summary>
		/// <param name="b"></param>
		/// <param name="buffer"></param>
		/// <param name="startIndex"></param>
		/// <param name="instanceNum"></param>
		/// <returns></returns>
		public static unsafe int IndexOfByCount(byte b, byte[] buffer, int startIndex, int instanceNum)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (instanceNum < 1 || startIndex > buffer.Length)
				throw new IndexOutOfRangeException();

			int count = 0;
			int length = buffer.Length;
			fixed (byte* arr = buffer)
			{
				byte* aptr = arr + startIndex;
				for (int i = startIndex; i < length; i++)
				{
					//if(aptr->Equals(b))
					if ((*aptr) == b)
					{
						++count;
						if(count == instanceNum)
							return i;
					}
					++aptr;
				}
			}
			return -1;
		}

		public static unsafe long LastIndexOf(byte b, byte[] buffer)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			long length = buffer.LongLength;
			fixed(byte* arr = buffer)
			{
				byte* aptr = arr;
				aptr += length;
				for(long i=length-1;i>-1;i--)
				{
					aptr--;
					if(aptr->Equals(b))
						return i;
				}
			}
			return -1;
		}

		/// <summary>
		/// The size of an Int32
		/// </summary>
		/// <value>4 (bytes)</value>
		public const int SizeOfInt32 = sizeof(int);

		public static unsafe void ToNetworkBytes(int valuetype, byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			// check for buffer underrun
			if (index < 0)
				throw new IndexOutOfRangeException();

			// check for buffer overrun
			// make sure to use the minus to prevent Integer overflow
			if (index > buffer.Length - SizeOfInt32)
				throw new IndexOutOfRangeException();

			fixed (byte* outarr = buffer)
			{
				byte* p = outarr + index;

				// do the swap copy
				swap_copy((byte*)&valuetype, p, SizeOfInt32);
			}
			//index += SizeOfInt32;
		}

		/// <summary>
		/// Size of the Double data type
		/// </summary>
		public const int SizeOfDouble = sizeof(double);
		/// <summary>
		/// Copies the specified double valuetype into the byte array using network byte order
		/// </summary>
		/// <param name="valuetype">The Double to convert</param>
		/// <param name="buffer">The buffer in which to place the double</param>
		/// <param name="index">The index at which to start copying</param>
		public static unsafe void ToNetworkBytes(double valuetype, byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			// check for buffer underrun
			if (index < 0)
				throw new IndexOutOfRangeException();

			// check for buffer overrun
			if (index > buffer.Length - SizeOfDouble)
				throw new IndexOutOfRangeException();

			fixed (byte* outarr = buffer)
			{
				byte* p = outarr + index;

				// do the swap copy
				swap_copy((byte*)&valuetype, p, SizeOfDouble);
			}
			//index += SizeOfDouble;
		}

		/// <summary>
		/// Size of the Single (float) data type
		/// </summary>
		public const int SizeOfSingle = sizeof(float);
		/// <summary>
		/// Converts the specified float value into network byte order
		/// </summary>
		/// <param name="valuetype">float value</param>
		/// <param name="buffer">buffer</param>
		/// <param name="index">index to start copying</param>
		public static unsafe void ToNetworkBytes(float valuetype, byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			// check for buffer underrun
			if (index < 0)
				throw new IndexOutOfRangeException();

			// check for buffer overrun
			if (index > buffer.Length - SizeOfSingle)
				throw new IndexOutOfRangeException();

			fixed (byte* outarr = buffer)
			{
				byte* p = outarr + index;

				// do the swap copy
				swap_copy((byte*)&valuetype, p, SizeOfSingle);
			}
			//index += SizeOfSingle;
		}

		// Decimal Removed because I'm not sure if it's a fixed size or not
		// Sizeof(Decimal) can't be assigned to a const
		/*/// <summary>
		/// Converts the decimal into network byte order for transmission.
		/// </summary>
		/// <remarks><see cref="System.Decimal"/> is a 12-byte type which more precisely preserves data.
		/// This data type may not be portable.</remarks>
		/// <param name="dec"></param>
		/// <param name="buffer"></param>
		/// <param name="index"></param>
		public static unsafe void ToNetworkBytes(decimal valuetype, byte[] buffer, ref int index)
		{
			const int size = sizeof(decimal);
			// check for buffer underrun
			if (index < 0)
				throw new IndexOutOfRangeException();

			// check for buffer overrun
			if (index + size > buffer.Length)
				throw new IndexOutOfRangeException();

			fixed (byte* outarr = buffer)
			{
				byte* p = outarr + index;

				// do the swap copy
				swap_copy((byte*)&valuetype, p, size);
			}
			index += size;
		}*/

		/// <summary>
		/// Size of an Int16
		/// </summary>
		/// <value>2 (bytes)</value>
		public const int SizeOfInt16 = sizeof(short);
		/// <summary>
		/// Convert the specified 16-bit signed integer to its network byte-order representation
		/// </summary>
		/// <param name="valuetype">The 16-bit integer to convert</param>
		/// <param name="buffer">The array into which the short should be copied</param>
		/// <param name="index">Index at which to start copying</param>
		public static unsafe void ToNetworkBytes(short valuetype, byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			// check for buffer underrun
			if (index < 0)
				throw new IndexOutOfRangeException();

			// check for buffer overrun
			if (index > buffer.Length - SizeOfInt16)
				throw new IndexOutOfRangeException();

			fixed (byte* outarr = buffer)
			{
				byte* p = outarr + index;

				// do the swap copy
				swap_copy((byte*)&valuetype, p, SizeOfInt16);
			}
			//index += SizeOfInt16;
		}

		/// <summary>
		/// Size of UInt16 (ushort)
		/// </summary>
		public const int SizeOfUInt16 = sizeof(ushort);
		/// <summary>
		/// Convert the specified 16-bit unsigned integer to its network byte-order representation
		/// </summary>
		/// <param name="valuetype">The 16-bit integer to convert</param>
		/// <param name="buffer">Buffer to copy into</param>
		/// <param name="index">Index at which to start copying</param>
		/// <returns>The byte array representation of the integer</returns>
		[CLSCompliant(false)]
		public static unsafe void ToNetworkBytes(ushort valuetype, byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			// check for buffer underrun
			if (index < 0)
				throw new IndexOutOfRangeException();

			// check for buffer overrun
			if (index > buffer.Length - SizeOfUInt16)
				throw new IndexOutOfRangeException();

			fixed (byte* outarr = buffer)
			{
				byte* p = outarr + index;

				// do the swap copy
				swap_copy((byte*)&valuetype, p, SizeOfUInt16);
			}
			//index += SizeOfUInt16;
		}

		/// <summary>
		/// Size of UInt32 in bytes
		/// </summary>
		public const int SizeOfUInt32 = sizeof(uint);
		/// <summary>
		/// Convert the specified 32-bit unsigned integer to its network byte-order representation
		/// </summary>
		/// <param name="valuetype">The 32-bit integer to convert</param>
		/// <param name="buffer">Buffer to copy into</param>
		/// <param name="index">Index to start copying</param>
		/// <returns>The byte array representation of the integer</returns>
		[CLSCompliant(false)]
		public static unsafe void ToNetworkBytes(uint valuetype, byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			// check for buffer underrun
			if (index < 0)
				throw new IndexOutOfRangeException();

			// check for buffer overrun
			if (index > buffer.Length - SizeOfUInt32)
				throw new IndexOutOfRangeException();

			fixed (byte* outarr = buffer)
			{
				byte* p = outarr + index;

				// do the swap copy
				swap_copy((byte*)&valuetype, p, SizeOfUInt32);
			}
			//index += SizeOfUInt32;
		}

		/// <summary>
		/// Size of Int64 in bytes
		/// </summary>
		public const int SizeOfInt64 = sizeof(long);
		/// <summary>
		/// Convert the specified 64-bit signed integer to its network byte-order representation
		/// </summary>
		/// <param name="valuetype">The 64-bit integer to convert</param>
		/// <param name="buffer">Buffer to copy into</param>
		/// <param name="index">Index to start copying</param>
		public static unsafe void ToNetworkBytes(long valuetype, byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			// check for buffer underrun
			if (index < 0)
				throw new IndexOutOfRangeException();

			// check for buffer overrun
			if (index > buffer.Length - SizeOfInt64)
				throw new IndexOutOfRangeException();

			fixed (byte* outarr = buffer)
			{
				byte* p = outarr + index;

				// do the swap copy
				swap_copy((byte*)&valuetype, p, SizeOfInt64);
			}
			//index += SizeOfInt64;
		}

		/// <summary>
		/// Size of UInt64 in bytes
		/// </summary>
		public const int SizeOfUInt64 = sizeof(ulong);
		/// <summary>
		/// Convert the specified 64-bit unsigned integer to its network byte-order representation
		/// </summary>
		/// <param name="valuetype">The 64-bit integer to convert</param>
		/// <param name="buffer">Buffer to copy into</param>
		/// <param name="index">Index to start copying</param>
		[CLSCompliant(false)]
		public static unsafe void ToNetworkBytes(ulong valuetype, byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			// check for buffer underrun
			if (index < 0)
				throw new IndexOutOfRangeException();

			// check for buffer overrun
			if (index > buffer.Length - SizeOfUInt64)
				throw new IndexOutOfRangeException();

			fixed (byte* outarr = buffer)
			{
				byte* p = outarr + index;

				// do the swap copy
				swap_copy((byte*)&valuetype, p, SizeOfUInt64);
			}
			//index += SizeOfUInt64;
		}

		/// <summary>
		/// Size of DateTime for transmission over network
		/// </summary>
		public const int SizeOfDateTime = SizeOfInt64;

		/*/// <summary>
		/// Converts DateTime to binary representation
		/// </summary>
		/// <param name="dateTime"></param>
		/// <param name="buffer"></param>
		/// <param name="index"></param>
		/// <remarks>New in .NET 2.0, a DateTime can be converted to its binary
		/// equivalent using DateTime.ToBinary()</remarks>
		public static void ToNetworkBytes(DateTime dateTime, byte[] buffer, int index)
		{
			ToNetworkBytes(dateTime.ToBinary(), buffer, index);
		}*/

		/// <summary>
		/// Size of the Guid structure
		/// </summary>
		/// <value>16 (bytes)</value>
		public const int SizeOfGUID = 16;
		/// <summary>
		/// Converts the Guid into a byte array
		/// </summary>
		/// <remarks><see cref="System.Guid"/> is a 16-byte structure used as an identifier.  This method does not
		/// change any byte ordering.  While the native GUID structure on Windows may use smaller types, 
		/// .NET implementations outside Windows (Mono) seem to only use a byte[16] internally</remarks>
		/// <param name="id">The GUID Structure to convert</param>
		/// <param name="buffer"></param>
		/// <param name="index"></param>
		public static unsafe void ToNetworkBytes(Guid id, byte[] buffer, int index)
		{
			/*fixed (byte* outarr = buffer)
			{
				byte* p = outarr;
				p += index;
				
			}*/
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (index < 0)
				throw new IndexOutOfRangeException();

			if (index > buffer.Length - SizeOfGUID)
				throw new IndexOutOfRangeException();

			fixed (byte* outarr = buffer)
			{
				copy4((byte*)&id, 0, outarr, index, SizeOfGUID);
			}
			//index += SizeOfGUID;
		}

		/// <summary>
		/// Converts the next 16 bytes in the array into a Guid
		/// </summary>
		/// <remarks><see cref="System.Guid"/> contains .ToByteArray() and can take a byte array as a constructor. 
		/// However, this requires the array to be a fixed size.  This method works on a buffer of any size.</remarks>
		/// <param name="buffer">The array that contains the GUID</param>
		/// <param name="index">Index at which to start copying</param>
		/// <returns>The Guid structure represented by this instance</returns>
		public static unsafe Guid ToGUID(byte[] buffer, ref int index)
		{
			/*Guid test = new Guid(ByteConverter.ToUInt32(buffer, index),
				ByteConverter.ToUInt16(buffer, index + 4), ByteConverter.ToUInt16(buffer, index + 6),
				buffer[index + 8], buffer[index + 9], buffer[index + 10], buffer[index + 11],
				buffer[index + 12], buffer[index + 13], buffer[index + 14], buffer[index + 15]);
			return test;*/
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (index < 0)
				throw new IndexOutOfRangeException();

			if (index > buffer.Length - SizeOfGUID)
				throw new IndexOutOfRangeException();

			Guid test;
			fixed (byte* inarr = buffer)
			{
				copy4(inarr, index, (byte*)&test, 0, SizeOfGUID);
			}
			index += SizeOfGUID;
			return test;
		}

		/*/// <summary>
		/// Converts the specified index in the byte array to a DateTime using the Binary representation
		/// </summary>
		/// <param name="array">Array at which to extract info</param>
		/// <param name="index">Index to start extracting</param>
		/// <returns></returns>
		public static DateTime ToDateTime(byte[] array, ref int index)
		{
			return DateTime.FromBinary(ToInt64(array, ref index));
		}*/

		/// <summary>
		/// Returns the 16-bit unsigned integer that is represented by the specified position in the array,
		/// using network byte-order representation
		/// </summary>
		/// <param name="buffer">The array to take the integer from</param>
		/// <param name="index">The index to start copying from</param>
		/// <returns>A 16-bit unsigned integer matching the bytes in the array</returns>
		[CLSCompliant(false)]
		public static unsafe ushort ToUInt16(byte[] buffer, ref int index)
		{
			ushort returnval = ToUInt16(buffer,index);
			index += SizeOfUInt16;
			return returnval;
		}

		/// <summary>
		/// Returns the 16-bit unsigned integer that is represented by the specified position in the array,
		/// using network byte-order representation
		/// </summary>
		/// <param name="buffer">The array to take the integer from</param>
		/// <param name="index">The index to start copying from</param>
		/// <returns>A 16-bit unsigned integer matching the bytes in the array</returns>
		[CLSCompliant(false)]
		public static unsafe ushort ToUInt16(byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if(index < 0)
				throw new IndexOutOfRangeException();

			if (index > buffer.Length - SizeOfUInt16)
				throw new IndexOutOfRangeException();

			ushort returnval;
			fixed(byte* inarr = buffer)
			{
				byte* ip = inarr;
				ip += index; // move pointer to index
				swap_copy(ip,(byte*)&returnval,SizeOfUInt16);
			}
			return returnval;
		}

		/// <summary>
		/// Returns the 16-bit signed integer that is represented by the specified position in the array,
		/// using network byte-order representation
		/// </summary>
		/// <param name="buffer">The array to take the integer from</param>
		/// <param name="index">The index to start copying from</param>
		/// <returns>A 16-bit signed integer matching the bytes in the array</returns>
		public static unsafe short ToInt16(byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (index < 0)
				throw new IndexOutOfRangeException();

			if (index > buffer.Length - SizeOfInt16)
				throw new IndexOutOfRangeException();

			short returnval;
			fixed(byte* inarr = buffer)
			{
				byte* ip = inarr + index;
				swap_copy(ip,(byte*)&returnval,SizeOfInt16);
			}
			index += SizeOfInt16;
			return returnval;
		}

		/// <summary>
		/// Returns the 16-bit signed integer that is represented by the specified position in the array,
		/// using network byte-order representation
		/// </summary>
		/// <param name="buffer">The array to take the integer from</param>
		/// <param name="index">The index to start copying from</param>
		/// <returns>A 16-bit signed integer matching the bytes in the array</returns>
		public static unsafe short ToInt16(byte[] buffer, ref int index)
		{
			short returnval= ToInt16(buffer,index);
			index += SizeOfInt16;
			return returnval;
		}

		/// <summary>
		/// Returns the 32-bit signed integer that is represented by the specified position in the array,
		/// using network byte-order representation
		/// </summary>
		/// <param name="buffer">The array to take the integer from</param>
		/// <param name="index">The index to start copying from</param>
		/// <returns>A 32-bit signed integer matching the bytes in the array</returns>
		public static unsafe int ToInt32(byte[] buffer, ref int index)
		{
			int returnval = ToInt32(buffer,index);
			index += SizeOfInt32;
			return returnval;
		}

		/// <summary>
		/// Returns the 32-bit signed integer that is represented by the specified position in the array,
		/// using network byte-order representation
		/// </summary>
		/// <param name="buffer">The array to take the integer from</param>
		/// <param name="index">The index to start copying from</param>
		/// <returns>A 32-bit signed integer matching the bytes in the array</returns>
		public static unsafe int ToInt32(byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (index < 0)
				throw new IndexOutOfRangeException();

			if (index > buffer.Length - SizeOfInt32)
				throw new IndexOutOfRangeException();

			int returnval;
			fixed(byte* inarr = buffer)
			{
				byte* ip = inarr + index;
				swap_copy(ip,(byte*)&returnval,SizeOfInt32);
			}
			//index += SizeOfInt32;
			return returnval;
		}

		/// <summary>
		/// Returns the 32-bit unsigned integer that is represented by the specified position in the array,
		/// using network byte-order representation
		/// </summary>
		/// <param name="buffer">The array to take the integer from</param>
		/// <param name="index">The index to start copying from</param>
		/// <returns>A 32-bit unsigned integer matching the bytes in the array</returns>
		[CLSCompliant(false)]
		public static unsafe uint ToUInt32(byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (index < 0)
				throw new IndexOutOfRangeException();

			if (index > buffer.Length - SizeOfUInt32)
				throw new IndexOutOfRangeException();

			uint returnval;
			fixed(byte* inarr = buffer)
			{
				byte* ip = inarr;
				ip += index; // move pointer to index
				swap_copy(ip,(byte*)&returnval,4);
			}
			return returnval;
		}

		/// <summary>
		/// Converts the byte array into a single (4-byte floating point number)
		/// </summary>
		/// <remarks>Floating point numbers may not be stored the same way
		/// in different architectures (aside from endianness).  Use caution
		/// with this method</remarks>
		/// <param name="buffer">Array in which the Single is contained</param>
		/// <param name="index">Start index where the array is</param>
		/// <returns>The <see cref="System.Single"/> the array contains</returns>
		/// <exception cref="System.IndexOutOfRangeException"/>
		public static unsafe float ToSingle(byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (index < 0)
				throw new IndexOutOfRangeException();

			if (index > buffer.Length - SizeOfSingle)
				throw new IndexOutOfRangeException();

			float returnval;
			fixed(byte* inarr = buffer)
			{
				byte* ip = inarr;
				ip += index; // move pointer to index
				swap_copy(ip,(byte*)&returnval,4);
			}
			return returnval;
		}

		/// <summary>
		/// Returns the 64-bit signed integer that is represented by the specified position in the array,
		/// using network byte-order representation
		/// </summary>
		/// <param name="buffer">The array to take the integer from</param>
		/// <param name="index">The index to start copying from</param>
		/// <returns>A 64-bit signed integer matching the bytes in the array</returns>
		/// <exception cref="System.IndexOutOfRangeException">Occures when
		/// index + sizeof(Int64) is greater than the length of the array
		/// or when index is less than 0</exception>
		public static unsafe long ToInt64(byte[] buffer, ref int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (index < 0)
				throw new IndexOutOfRangeException();

			if (index > buffer.Length - SizeOfInt64)
				throw new IndexOutOfRangeException();

			long returnval;
			fixed(byte* inarr = buffer)
			{
				byte* ip = inarr;
				ip += index; // move pointer to index
				swap_copy(ip,(byte*)&returnval,8);
			}
			index += SizeOfInt64;
			return returnval;
		}

		/// <summary>
		/// Returns the 64-bit unsigned integer that is represented by the specified position in the array,
		/// using network byte-order representation
		/// </summary>
		/// <param name="buffer">The array to take the integer from</param>
		/// <param name="index">The index to start copying from</param>
		/// <returns>A 64-bit unsigned integer matching the bytes in the array</returns>
		[CLSCompliant(false)]
		public static unsafe ulong ToUInt64(byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (index < 0)
				throw new IndexOutOfRangeException();

			if (index > buffer.Length - SizeOfUInt64)
				throw new IndexOutOfRangeException();

			ulong returnval;
			fixed(byte* inarr = buffer)
			{
				byte* ip = inarr;
				ip += index; // move pointer to index
				swap_copy(ip,(byte*)&returnval,8);
			}
			return returnval;
		}

		/// <summary>
		/// Converts the arary into a double
		/// </summary>
		/// <remarks><see cref="System.Double"/> may not be cross platform
		/// even with byte swapping.  Testing is required</remarks>
		/// <param name="buffer">The array in which the double is stored</param>
		/// <param name="index">Index in the array at which the double starts</param>
		/// <returns>The double value</returns>
		public static unsafe double ToDouble(byte[] buffer, int index)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (index < 0)
				throw new IndexOutOfRangeException();

			if (index > buffer.Length - SizeOfDouble)
				throw new IndexOutOfRangeException();

			double returnval;
			fixed(byte* inarr = buffer)
			{
				byte* ip = inarr;
				ip += index; // move pointer to index
				swap_copy(ip,(byte*)&returnval,8);
			}
			return returnval;
		}

		// I really wish this worked, but I get CS0702
		/*public static unsafe T ToType<T>(byte[] array, int index)
			where T : System.ValueType
		{
			T returnval;
			fixed (byte* inarr = array)
			{
				byte* ip = inarr;
				ip += index;
				swap_copy(ip, (byte*)&returnval, sizeof(T));
			}
			return returnval;
		}*/

		/// <summary>
		/// Return a subset of the specified array
		/// </summary>
		/// <param name="buffer">The array to pull data from</param>
		/// <param name="startIndex">The index to start copying data</param>
		/// <param name="length">The length of the new array</param>
		/// <returns>A new byte array with data from the specified portion of the original array</returns>
		public static unsafe byte[] SubArray(byte[] buffer, int startIndex, int length)
		{
			// these checks are performed by Copy
			//if(index < 0 || array.Length - index < 0 || array.Length - index - length < 0)
			//	throw new ArgumentException();
			/*if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (startIndex < 0)
				throw new IndexOutOfRangeException();

			if (length < 0)
				throw new ArgumentOutOfRangeException("length");

			if (buffer.Length < startIndex)
				throw new ArgumentOutOfRangeException("index", "Index is bigger than the array");

			if (startIndex > buffer.Length - length)
				throw new IndexOutOfRangeException();*/

			byte[] returnval = new byte[length];

			/*fixed(byte* inarr = array, outarr = returnval)
			{
				byte* ps = inarr;
				ps += index; // move to index
				byte* pd = outarr;*/

			/*if(useNetworkByteOrder)
			{
				swap_copy(ps,pd,length);
			}
			else
			{*/
			Copy(buffer, startIndex, returnval, 0, length);
			//}
			//}

			return returnval;
		}

		/// <summary>
		/// Copies data from one byte array to another, using pointers
		/// </summary>
		/// <param name="src">The source byte array</param>
		/// <param name="srcIndex">The 0-based index in the source array to start copying from</param>
		/// <param name="dst">The destination array</param>
		/// <param name="dstIndex">The 0-based index in the destination array to start copying to</param>
		/// <param name="count">The number of bytes to copy</param>
		/// <exception cref="System.ArgumentException">Thrown when the arrays are null, or any of the parameters are negative,
		/// or when copying would exceed the lengths of the arrays</exception>
		public static unsafe void Copy(byte[] src, int srcIndex,
			byte[] dst, int dstIndex, int count)
		{
			if (count == 0) // just do nothing if count is 0
				return;
			if (src == null)
				throw new ArgumentNullException("src");
			if (dst == null)
				throw new ArgumentNullException("dst");
			if (srcIndex < 0 || dstIndex < 0)
				throw new IndexOutOfRangeException();
			if (count < 0)
				throw new ArgumentOutOfRangeException("count", count, "negative");

			int srcLen = src.Length;
			int dstLen = dst.Length;
			if (srcLen - srcIndex < count ||
				dstLen - dstIndex < count)
			{
				throw new ArgumentException("Source Length " + srcLen + ", Source Index " + srcIndex + "\r\n" +
					"Dest Length " + dstLen + ", Dest Index " + dstIndex + ", Count " + count);
			}

			// The following fixed statement pins the location of
			// the src and dst objects in memory so that they will
			// not be moved by garbage collection.          
			fixed (byte* pSrc = src, pDst = dst)
			{
				copy4(pSrc, srcIndex, pDst, dstIndex, count);
			}
		}

		/// <summary>
		/// Return a subset of the specified array
		/// </summary>
		/// <param name="buffer">The array to pull data from</param>
		/// <param name="index">The index to start copying data</param>
		/// <param name="length">The length of the new array</param>
		/// <returns>A new byte array with data from the specified portion of the original array</returns>
		public static unsafe byte[] SubArray(byte[] buffer, long index, long length)
		{
			// these checks are performed by Copy
			//if(index < 0 || array.Length - index < 0 || array.Length - index - length < 0)
			//	throw new ArgumentException();
			/*if (buffer == null)
				throw new ArgumentNullException("buffer");

			if (index < 0)
				throw new IndexOutOfRangeException();

			if (length < 0)
				throw new IndexOutOfRangeException();

			if (buffer.LongLength < index)
				throw new ArgumentOutOfRangeException("index", "Index is bigger than the array");

			if (index > buffer.LongLength - length)
				throw new IndexOutOfRangeException();*/

			byte[] returnval = new byte[length];

			/*fixed(byte* inarr = array, outarr = returnval)
			{
				byte* ps = inarr;
				ps += index; // move to index
				byte* pd = outarr;*/

				/*if(useNetworkByteOrder)
				{
					swap_copy(ps,pd,length);
				}
				else
				{*/
					Copy(buffer,index,returnval,0,length);
				//}
			//}

			return returnval;
		}

		/// <summary>
		/// Copies data from one byte array to another, using pointers
		/// </summary>
		/// <param name="src">The source byte array</param>
		/// <param name="srcIndex">The 0-based index in the source array to start copying from</param>
		/// <param name="dst">The destination array</param>
		/// <param name="dstIndex">The 0-based index in the destination array to start copying to</param>
		/// <param name="count">The number of bytes to copy</param>
		/// <exception cref="System.ArgumentException">Thrown when the arrays are null, or any of the parameters are negative,
		/// or when copying would exceed the lengths of the arrays</exception>
		public static unsafe void Copy(byte[] src, long srcIndex,
			byte[] dst, long dstIndex, long count)
		{
			if (count == 0) // just do nothing if count is 0
				return;
			if (src == null)
				throw new ArgumentNullException("src");
			if (dst == null)
				throw new ArgumentNullException("dst");
			if (srcIndex < 0 || dstIndex < 0)
				throw new IndexOutOfRangeException();
			if (count < 0)
				throw new ArgumentOutOfRangeException("count", count, "negative");

			long srcLen = src.LongLength;
			long dstLen = dst.LongLength;
			if (srcLen - srcIndex < count ||
				dstLen - dstIndex < count)
			{
				throw new ArgumentException("Source Length " + srcLen + ", Source Index " + srcIndex + "\r\n" +
					"Dest Length " + dstLen + ", Dest Index " + dstIndex + ", Count " + count);
			}
  
			// The following fixed statement pins the location of
			// the src and dst objects in memory so that they will
			// not be moved by garbage collection.          
			fixed (byte* pSrc = src, pDst = dst)
			{
				copy4(pSrc, srcIndex, pDst, dstIndex, count);
			}
		}

		/// <summary>
		/// Write 0 to specified areas of the byte array
		/// </summary>
		/// <param name="buffer">Buffer to zeroize</param>
		/// <param name="startIndex"></param>
		/// <param name="count"></param>
		/// <remarks>See http://pages.infinit.net/ctech/20050610-0319.html</remarks>
		public static unsafe void Zeroize(byte[] buffer, int startIndex, int count)
		{
			if (buffer == null)
				throw new ArgumentNullException("buffer");
			if (startIndex < 0)
				throw new IndexOutOfRangeException();
			if (count < 0)
				throw new ArgumentOutOfRangeException("count",count,"negative");
			if (startIndex > buffer.Length - count)
				throw new ArgumentOutOfRangeException("startIndex + count > this.length");

			int i = 0;
			fixed (byte* ptr = &buffer[startIndex])
			{
				byte* p = ptr;
				while (i++ < count)
					*p++ = 0x00;
			}
		}

		// performs byte copy in blocks of 4
		// this is faster than doing a byte-by-byte copy
		// is only useful when not worrying about endianness
		// i.e. to copy from one byte array to another
		private static unsafe void copy4(byte* inarr, int srcIndex, byte* outarr,
			int dstIndex, int count)
		{
			byte* ps = inarr;
			byte* pd = outarr;
			ps += srcIndex; // scroll ahead (Missing from example)
			pd += dstIndex; // scroll ahead (Missing from example)

			// Loop over the count in blocks of 4 bytes, copying an
			// integer (4 bytes) at a time:
			for (int n = 0; n < count / 4; n++)
			{
				*((int*)pd) = *((int*)ps);
				pd += 4;
				ps += 4;
			}

			// Complete the copy by moving any bytes that weren't
			// moved in blocks of 4:
			for (int n = 0; n < count % 4; n++)
			{
				*pd = *ps;
				pd++;
				ps++;
			}
		}

		// performs byte copy in blocks of 4
		// this is faster than doing a byte-by-byte copy
		// is only useful when not worrying about endianness
		// i.e. to copy from one byte array to another
		private static unsafe void copy4(byte* inarr, long srcIndex, byte* outarr,
			long dstIndex, long count)
		{
			byte* ps = inarr;
			byte* pd = outarr;
			ps += srcIndex; // scroll ahead (Missing from example)
			pd += dstIndex; // scroll ahead (Missing from example)

			// Loop over the count in blocks of 4 bytes, copying an
			// integer (4 bytes) at a time:
			for (int n = 0; n < count / 4; n++)
			{
				*((int*)pd) = *((int*)ps);
				pd += 4;
				ps += 4;
			}

			// Complete the copy by moving any bytes that weren't
			// moved in blocks of 4:
			for (int n = 0; n < count % 4; n++)
			{
				*pd = *ps;
				pd++;
				ps++;
			}
		}

		// stricly internal class only, used for doing the actual pointer manipulation
		private static unsafe void swap_copy(byte* inarr,byte* outarr,int len)
		{
			byte* op = outarr; // pointer to out array
			byte* ip = inarr; // pointer to in array

			// if we are in big endian, we preserve the byte ordering
			if(IsBigEndian)
			{
				for(int i=0;i<len;i++)
				{
					*op = *ip;
					op++;
					ip++;
				}
			}
			else // otherwise, go in one direction for the inarr, and another for the out arr
			{
				op += (len - 1); // "scroll" ahead for the out pointer
				for(int i=0;i<len;i++)
				{
					*op = *ip;
					op--;
					ip++;
				}
			}
			//}
			//return outarr;
		}
	}
}
